import * as i0 from '@angular/core';
import { Inject, Injectable, NgModule } from '@angular/core';
import { BehaviorSubject, Observable, from, Subject, combineLatest, of } from 'rxjs';
import cube, { moveItemInArray, isQueryPresent, movePivotItem, ResultSet, defaultHeuristics } from '@cubejs-client/core';
import { switchMap, catchError } from 'rxjs/operators';
import equal from 'fast-deep-equal/es6';

class CubeClient {
    config;
    ready$ = new BehaviorSubject(false);
    cubeApi;
    constructor(config) {
        this.config = config;
        if (this.config instanceof Observable) {
            this.config.subscribe(() => {
                this.ready$.next(true);
            });
        }
        else {
            this.ready$.next(true);
        }
    }
    apiInstance() {
        if (!this.cubeApi) {
            if (this.config instanceof Observable) {
                this.config.subscribe((config) => {
                    this.cubeApi = cube(config.token, config.options);
                    if (!this.cubeApi) {
                        throw new Error('Cannot create CubeApi instance. Please check that the config is passed correctly and contains all required options.');
                    }
                });
            }
            else {
                this.cubeApi = cube(this.config.token, this.config.options);
            }
        }
        return this.cubeApi;
    }
    load(query, options) {
        return from(this.apiInstance().load(query, options));
    }
    sql(query, options) {
        return from(this.apiInstance().sql(query, options));
    }
    dryRun(query, options) {
        return from(this.apiInstance().dryRun(query, options));
    }
    meta(options) {
        return from(this.apiInstance().meta(options));
    }
    watch(query, params = {}) {
        return new Observable((observer) => query.subscribe({
            next: async (query) => {
                try {
                    const resultSet = await this.apiInstance().load(query, params);
                    observer.next(resultSet);
                }
                catch (err) {
                    observer.error(err);
                }
            },
        }));
    }
    /** @nocollapse */ static ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: CubeClient, deps: [{ token: 'config' }], target: i0.ɵɵFactoryTarget.Injectable });
    /** @nocollapse */ static ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: CubeClient });
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: CubeClient, decorators: [{
            type: Injectable
        }], ctorParameters: () => [{ type: undefined, decorators: [{
                    type: Inject,
                    args: ['config']
                }] }] });

class CubeClientModule {
    static forRoot(config) {
        return {
            ngModule: CubeClientModule,
            providers: [
                CubeClient,
                {
                    provide: 'config',
                    useValue: config,
                },
            ],
        };
    }
    /** @nocollapse */ static ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: CubeClientModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
    /** @nocollapse */ static ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "18.2.13", ngImport: i0, type: CubeClientModule });
    /** @nocollapse */ static ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: CubeClientModule, providers: [CubeClient] });
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: CubeClientModule, decorators: [{
            type: NgModule,
            args: [{
                    providers: [CubeClient],
                }]
        }] });

class StateSubject {
    subject;
    constructor(value) {
        this.subject = new BehaviorSubject(value);
    }
    get() {
        return this.subject.getValue();
    }
    set(value) {
        this.subject.next(value);
    }
}

class BaseMember {
    query;
    field;
    constructor(query, field) {
        this.query = query;
        this.field = field;
    }
    get members() {
        return this.query.asCubeQuery()[this.field] || [];
    }
    add(name) {
        this.query.setPartialQuery({
            [this.field]: [...this.members, name],
        });
    }
    replace(name, replaceWithName) {
        this.query.setPartialQuery({
            [this.field]: this.members.map((currentName) => currentName === name ? replaceWithName : currentName),
        });
    }
    remove(by) {
        this.query.setPartialQuery({
            [this.field]: this.query
                .asCubeQuery()[this.field].filter((currentName, index) => {
                if (typeof by === 'string') {
                    return currentName !== by;
                }
                return index !== by;
            }),
        });
    }
    set(members) {
        this.query.setPartialQuery({
            [this.field]: members,
        });
    }
    asArray() {
        return (this.query.asCubeQuery()[this.field] || []).map((name) => this.query.meta.resolveMember(name, this.field));
    }
}
class TimeDimensionMember {
    query;
    constructor(query) {
        this.query = query;
    }
    get members() {
        return this.query.asCubeQuery().timeDimensions || [];
    }
    get granularity() {
        return this.members[0]?.granularity;
    }
    updateTimeDimension(by, updateWith) {
        const timeDimensions = this.members.map((td, index) => {
            if (td.dimension === by || index === by) {
                return {
                    ...td,
                    ...updateWith,
                };
            }
            return td;
        });
        this.query.setPartialQuery({
            timeDimensions,
        });
    }
    add(name) {
        this.query.setPartialQuery({
            timeDimensions: [
                {
                    dimension: name,
                },
            ],
        });
    }
    remove(name) {
        this.query.setPartialQuery({
            timeDimensions: this.members.filter(({ dimension }) => dimension !== name),
        });
    }
    set(timeDimensions) {
        this.query.setPartialQuery({
            timeDimensions,
        });
    }
    setDateRange(by, dateRange) {
        this.updateTimeDimension(by, { dateRange });
    }
    setGranularity(by, granularity) {
        this.updateTimeDimension(by, { granularity });
    }
    asArray() {
        return (this.query.asCubeQuery().timeDimensions || []).map((td) => {
            return {
                ...this.query.meta.resolveMember(td.dimension, 'dimensions'),
                ...td,
            };
        });
    }
}
class Order {
    query;
    orderMembers = new BehaviorSubject([]);
    constructor(query) {
        this.query = query;
        this.query.subject.subscribe(this.handleQueryChange.bind(this));
        this.orderMembers.subscribe(this.handleOrderMembersChange.bind(this));
    }
    handleOrderMembersChange(orderMembers) {
        const order = orderMembers
            .filter(({ order }) => order !== 'none')
            .reduce((memo, { id, order }) => ({ ...memo, [id]: order }), {});
        if (!equal(order, this.asObject())) {
            this.query.setPartialQuery({ order });
        }
    }
    handleQueryChange() {
        this.orderMembers.next([
            ...this.query.measures.asArray(),
            ...this.query.dimensions.asArray(),
            ...this.query.timeDimensions.asArray(),
        ].map(({ name, title }) => {
            return {
                id: name,
                order: this.of(name),
                title,
            };
        }));
    }
    setMemberOrder(id, order) {
        this.orderMembers.next(this.orderMembers.getValue().map((orderMember) => {
            if (orderMember.id === id) {
                return {
                    ...orderMember,
                    order,
                };
            }
            return orderMember;
        }));
    }
    reorder(sourceIndex, destinationIndex) {
        this.orderMembers.next(moveItemInArray(this.orderMembers.getValue(), sourceIndex, destinationIndex));
    }
    of(member) {
        return (this.query.asCubeQuery().order || {})[member] || 'none';
    }
    set(order) {
        this.query.setPartialQuery({ order });
    }
    asArray() {
        if (Array.isArray(this.query.asCubeQuery().order)) {
            return this.query.asCubeQuery().order;
        }
        return Object.entries(this.query.asCubeQuery().order || {});
    }
    asObject() {
        return this.asArray().reduce((memo, [key, value]) => ({ ...memo, [key]: value }), {});
    }
}
class FilterMember {
    query;
    constructor(query) {
        this.query = query;
    }
    get filters() {
        // TODO: update this type assertion once the QueryBuilder supports logical and/or
        return (this.query.asCubeQuery().filters || []);
    }
    update(by, updateWith) {
        const filters = this.filters.map((filter, index) => {
            if (index === by || filter.member === by || filter.dimension === by) {
                return {
                    ...filter,
                    ...updateWith,
                };
            }
            return filter;
        });
        this.query.setPartialQuery({
            filters: filters,
        });
    }
    add(filter) {
        this.query.setPartialQuery({
            filters: [...this.filters, filter],
        });
    }
    remove(by) {
        this.query.setPartialQuery({
            filters: this.filters.filter((filter, index) => {
                if (filter.member === by || filter.dimension === by || index === by) {
                    return false;
                }
                return true;
            }),
        });
    }
    set(filters) {
        this.query.setPartialQuery({
            filters,
        });
    }
    replace(name, replaceWithName) {
        this.query.setPartialQuery({
            filters: this.filters.map((filter) => {
                const field = filter.member ? 'member' : 'dimension';
                return filter.member === name || filter.dimension === name
                    ? {
                        ...filter,
                        [field]: replaceWithName,
                    }
                    : filter;
            }),
        });
    }
    asArray() {
        return this.filters.map((filter) => {
            return {
                ...this.query.meta.resolveMember(filter.member || filter.dimension, [
                    'dimensions',
                    'measures',
                ]),
                operators: this.query.meta.filterOperatorsForMember(filter.member || filter.dimension, ['dimensions', 'measures']),
                ...filter,
            };
        });
    }
}

var MemberType;
(function (MemberType) {
    MemberType["Measures"] = "measures";
    MemberType["Dimensions"] = "dimensions";
    MemberType["Segments"] = "segments";
    MemberType["TimeDimensions"] = "timeDimensions";
    MemberType["Filters"] = "filters";
    MemberType["Order"] = "order";
})(MemberType || (MemberType = {}));
class Query extends StateSubject {
    meta;
    _onBeforeChange;
    measures;
    dimensions;
    segments;
    timeDimensions;
    filters;
    order;
    constructor(meta, _onBeforeChange = (newQuery) => newQuery) {
        super({});
        this.meta = meta;
        this._onBeforeChange = _onBeforeChange;
        this.init();
    }
    init() {
        this.measures = new BaseMember(this, MemberType.Measures);
        this.dimensions = new BaseMember(this, MemberType.Dimensions);
        this.segments = new BaseMember(this, MemberType.Segments);
        this.timeDimensions = new TimeDimensionMember(this);
        this.filters = new FilterMember(this);
        this.order = new Order(this);
    }
    asCubeQuery() {
        return this.subject.getValue() || {};
    }
    setQuery(query) {
        this.subject.next(this._onBeforeChange(query, this.subject.getValue(), this));
    }
    setPartialQuery(partialQuery) {
        this.subject.next(this._onBeforeChange({
            ...this.subject.getValue(),
            ...partialQuery,
        }, this.subject.getValue(), this));
    }
    setLimit(limit) {
        this.setPartialQuery({ limit });
    }
    isPresent() {
        return isQueryPresent(this.asCubeQuery());
    }
}

class BuilderMeta {
    meta;
    measures;
    dimensions;
    segments;
    timeDimensions;
    filters;
    constructor(meta) {
        this.meta = meta;
        this.mapMeta();
    }
    mapMeta() {
        const allDimensions = (this.meta.membersForQuery(null, 'dimensions'));
        this.measures = this.meta.membersForQuery(null, 'measures');
        this.segments = this.meta.membersForQuery(null, 'segments');
        this.dimensions = allDimensions.filter(({ type }) => type !== 'time');
        this.timeDimensions = allDimensions.filter(({ type }) => type === 'time');
        this.filters = [...allDimensions, ...this.measures].map((member) => {
            return {
                ...member,
                operators: this.meta.filterOperatorsForMember(member.name, [
                    'dimensions',
                    'measures',
                ]),
            };
        });
    }
}

class PivotConfig extends StateSubject {
    constructor(pivotConfig) {
        super(pivotConfig);
    }
    moveItem(sourceIndex, destinationIndex, sourceAxis, destinationAxis) {
        this.subject.next(movePivotItem(this.get(), sourceIndex, destinationIndex, sourceAxis, destinationAxis));
    }
    setFillMissingDates(fillMissingDates) {
        this.subject.next({
            ...this.get(),
            fillMissingDates
        });
    }
}

class ChartType extends StateSubject {
    constructor(value) {
        super(value);
    }
}

class QueryBuilderService {
    _cube;
    _meta;
    _query;
    _disableHeuristics = false;
    _resolveQuery;
    _resolveBuilderMeta;
    _heuristicChange$ = new Subject();
    builderMeta = new Promise((resolve) => (this._resolveBuilderMeta = resolve));
    query = new Promise((resolve) => (this._resolveQuery = resolve));
    state = new BehaviorSubject({});
    pivotConfig;
    chartType;
    async init() {
        this.pivotConfig = new PivotConfig(null);
        this.chartType = new ChartType('line');
        this._cube.meta().subscribe((meta) => {
            this._meta = meta;
            this._query = new Query(this._meta, this._handleQueryChange.bind(this));
            this._resolveQuery(this._query);
            this._resolveBuilderMeta(new BuilderMeta(this._meta));
        });
        this.subscribe();
        if (!this._disableHeuristics) {
            this._heuristicChange$
                .pipe(switchMap((data) => {
                return combineLatest([
                    this._cube.dryRun(data.query).pipe(catchError((error) => {
                        console.error(error);
                        return of(null);
                    })),
                    of(data.shouldApplyHeuristicOrder),
                ]);
            }))
                .subscribe(([dryRunResponse, shouldApplyHeuristicOrder]) => {
                if (!dryRunResponse) {
                    return;
                }
                const { pivotQuery, queryOrder } = dryRunResponse;
                this.pivotConfig.set(ResultSet.getNormalizedPivotConfig(pivotQuery, this.pivotConfig.get()));
                if (shouldApplyHeuristicOrder) {
                    this._query.order.set(queryOrder.reduce((a, b) => ({ ...a, ...b }), {}));
                }
            });
        }
    }
    _handleQueryChange(newQuery, oldQuery) {
        const { chartType, shouldApplyHeuristicOrder, query: heuristicQuery, } = defaultHeuristics(newQuery, oldQuery, {
            meta: this._meta,
            sessionGranularity: newQuery?.timeDimensions?.[0]?.granularity,
        });
        const query = this._disableHeuristics
            ? newQuery
            : heuristicQuery || newQuery;
        if (isQueryPresent(query) && !this._disableHeuristics) {
            this._heuristicChange$.next({
                shouldApplyHeuristicOrder: Boolean(shouldApplyHeuristicOrder),
                query,
            });
        }
        if (!this._disableHeuristics && chartType) {
            this.chartType.set(chartType);
        }
        return query;
    }
    setCubeClient(cubeClient) {
        this._cube = cubeClient;
        this.init();
    }
    subscribe() {
        Object.getOwnPropertyNames(this).forEach((key) => {
            if (this[key] instanceof StateSubject) {
                this[key].subject.subscribe((value) => this.setPartialState({
                    [key]: value,
                }));
            }
        });
        this.query.then((query) => {
            query.subject.subscribe((cubeQuery) => {
                this.setPartialState({
                    query: cubeQuery,
                });
            });
        });
    }
    async deserialize(state) {
        if (state.query) {
            (await this.query).setQuery(state.query);
        }
        Object.entries(state).forEach(([key, value]) => {
            if (this[key] instanceof StateSubject) {
                this[key].set(value);
            }
        });
        this.subscribe();
    }
    setPartialState(partialState) {
        this.state.next({
            ...this.state.getValue(),
            ...partialState,
        });
    }
    disableHeuristics() {
        this._disableHeuristics = false;
    }
    enableHeuristics() {
        this._disableHeuristics = true;
    }
    /** @nocollapse */ static ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: QueryBuilderService, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
    /** @nocollapse */ static ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: QueryBuilderService });
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "18.2.13", ngImport: i0, type: QueryBuilderService, decorators: [{
            type: Injectable
        }] });

/*
 * Public API Surface of cubejs-client-ngx
 */

// This file is not used to build this module. It is only used during editing
// by the TypeScript language service and during build for verification. `ngc`
// replaces this file with production index.ts when it rewrites private symbol
// names.

/**
 * Generated bundle index. Do not edit.
 */

export { BaseMember, BuilderMeta, ChartType, CubeClient, CubeClientModule, FilterMember, MemberType, Order, PivotConfig, Query, QueryBuilderService, TimeDimensionMember };
//# sourceMappingURL=cubejs-client-ngx.mjs.map
